<?php
/**
 * @file
 * Drupal Inspect module base class
 */

/**
 * Inspect base class
 *
 * @uses PHP:ini:precision
 * @static
 * @singleton
 * @category Drupal
 * @package Developer
 * @subpackage Inspect */
class Inspect {
  /**
   * Prevents erratic instantiation of this entirely static class
   */
  private function  __construct() {}

  /**
   * Maximum sub var recursion depth.
   *
   * @type integer
   */
  const DEPTH_MAX = 10;

  /**
   * Default sub var recursion depth.
   *
   * @type integer
   */
  const DEPTH_DEFAULT = 10;

  /**
   * Minimum string truncation.
   *
   * @type integer
   */
  const TRUNCATE_MIN = 10000;

  /**
   * Default string truncation.
   *
   * @type integer
   */
  const TRUNCATE_DEFAULT = 1000;

  /**
   * Maximum output length when logging.
   *
   * @var integer $_outputMax_log
   */
  protected static $_outputMax_log = 0;

  /**
   * Maximum output length when getting.
   *
   * @var integer $_outputMax_get
   */
  protected static $_outputMax_get = 0;

  /**
   * Maximum output length for current mode (log or get).
   *
   * @var integer $_outputMax
   */
  protected static $_outputMax = 0;

  /**
   * Times _inspect() has been called.
   *
   * @var integer $_callNo
   */
  protected static $_callNo = 0;

  /**
   * PHP max_execution_time as measured at first call to _inspect().
   *
   * @var integer $_maxExecTime
   */
  protected static $_maxExecTime = 0;

  /**
   * _inspect() aborts when called this percent of max_execution_time has passed.
   *
   * @var integer $_maxExecTimePercent
   */
  protected static $_maxExecTimePercent = 0;

  /**
   * _inspect() will abort if called later than this time ($_maxExecTimePercent percent of max execution time).
   *
   * @var integer $_maxExecTimeout
   */
  protected static $_maxExecTimeout = 0;

  /**
   * Base 62 uniqid() if session started, otherwise empty.
   *
   * @var string $_sessionNo
   */
  protected static $_sessionNo = '';

  /**
   * Request number in current session.
   *
   * @var string $_requestNo
   */
  protected static $_requestNo = 0;

  /**
   * Log number in current request.
   *
   * @var integer $_logNo
   */
  protected static $_logNo = 0;

  /**
   * Request time start in milliseconds.
   *
   * @var float $_requestTimeMilli
   */
  protected static $_requestTimeMilli = 0;

  /**
   * private://module/inspect
   *
   * @var string|NULL|boolean $_dir
   */
  protected static $_dir;

  /**
   * @var string|NULL $_logDir
   */
  protected static $_logDir;

  /**
   * Default inspection options.
   *
   *   Various options:
   *   - (array) category: logging category
   *   - (boolean) file_line: display file and line of call to inspect function/method
   *   - (array|NULL) filter: filter out that|those key name(s)
   *   - (boolean) one_lined: for creating one-lined formatting instead of multi-lined
   *
   *   String options:
   *   - (array) needles: list of char|string needles to be used in str_replace
   *   - (array) replacers: list of char|string replacers to be used in str_replace
   *   - (string) quote_begin: char|string used as initial quote
   *   - (string) quote_end: char|string used as final quote
   *
   *   Formatting options:
   *   - (array) delimiters: between object/array buckets, and for InspectTrace between frames and frame buckets
   *   - (string) pre_indent: indentation of object/array buckets
   *   - (string) indent: indentation of object/array buckets (gets multiplied by current depth)
   *   - (string) enclose_tag: html enclosure tag
   *   - (string) newline: newline character
   *
   * @var array $_options
   */
  protected static $_options = array(
    'category' => 'inspect',
    'file_line' => TRUE,
    'filter' => NULL,
    'one_lined' => FALSE,
    'needles' => array( "\n", "\r", "\t", '<', '>', '"', "'" ),
    'replacers' => array( '_NL_', '_CR_', '_TB_', '&#60;', '&#62;', '&#34;', '&#39;' ),
    'quote_begin' => '&#62;&#62;',
    'quote_end' => '&#60;&#60;',
    'delimiters' => array(
        "\n", // Before first bucket.
        "\n", // Before later bucket.
        "\n"  // Before last bucket.
    ),
    'pre_indent' => '',
    'indent' => '.  ',
    'enclose_tag' => 'pre',
    'newline' => "\n",
  );

  /**
   * Options overriding the default, when target is file or mode is CLI/Drush and target is get.
   *
   * @var array $_options_file
   */
  protected static $_options_file = array(
    'quote_begin' => '"',
    'quote_end' => '"',
    'enclose_tag' => '',
  );

  /**
   * Options overriding the default, when truthy option one_lined.
   *
   * Apart form formatting, also makes default depth 1 and call file and line default off.
   *
   * @var array $_options_oneLined
   */
  protected static $_options_oneLined = array(
    'file_line' => FALSE,
    'delimiters' => array('', ', ', ''),
    'indent' => '',
    'enclose_tag' => '',
    'newline' => ' ',
    'no_preface' => TRUE,
    'name' => 'none',
  );

  /**
   * Whether server OS is *nix (non-windows, really).
   *
   * @var integer $_os_nix
   */
  protected static $_os_nix = -1; // ~ never checked

  /**
   * Called via command line (Drush)?
   *
   * @var boolean $_isCli
   */
  protected static $_isCli = FALSE;

  /**
   * @var integer|string $_userId
   */
  protected static $_userId;

  /**
   * Whether the PHP multibyte extension exists and is loaded.
   *
   * @var integer $_mbString_exists
   */
  protected static $_mbString_exists = -1; // ~ never checked

  /**
   * Document root and - if symlinked root real path (__FILE__ vs. SCRIPT_FILENAME) - with trailing slash, and as regular expression(s) (/^.../)
   *
   * @var array $_paths
   */
  protected static $_paths;

  /**
   * Length of document root or real path (the shorter one of them).
   *
   * @var integer $_pathLength
   */
  protected static $_pathLength;

  /**
   * Cache of the module permissions.
   *
   * @var array $_permissions
   */
  private static $_permissions = array(
    'inspect log' => FALSE,
    'inspect get' => FALSE,
  );

  /**
   * String length of an inspection output a any time during inspection.
   *
   * @var integer $_outputLength
   */
  protected static $_outputLength = 0;

  /**
   * _inspect() returns empty if this isnt truthy.
   *
   * @var bool $_initialized
   */
  private static $_initialized = FALSE;

  /**
   * Establish maximum output length, based on hardcoded absolute max and the database's max query length (or default for database).
   *
   * If $init and no persistent variables values: sets new values.
   *
   * Maximum for filing is the same as max for getting.
   *
   * Public because module settings dialog calls this method.
   *
   * @param boolean $init
   *  - default: FALSE (~ returns array { logDefault: N0, log: N1, get: N2 }
   *  - truthy: simply loads the values from persistent variables
   * @return boolean|array
   */
  public static function outputMax($init = FALSE) {
    if ($init
        && (self::$_outputMax_log = variable_get('inspect_outputmax_log', 0)) && (self::$_outputMax_get = variable_get('inspect_outputmax_get', 0))) {
      return TRUE;
    }
    $absolute = 2097152; // 2 Mb
    $logDefault = 1048576; // 1 Mb ~ MySql default max_allowed_packet
    //  Find max string length of db queries (MySql only).
    try {
      if (!($log = (int) db_query('SELECT @@max_allowed_packet')->fetchField())) {
        $log = $logDefault;
      }
      elseif ($log > $absolute) {
        $log = $absolute;
      }
    }
    catch (Exception $xc) { // Possible PDOException.
      self::log_db('Failed to read max query length of database', 'inspect', WATCHDOG_WARNING);
      $log = $logDefault;
    }
    //  Max watchdog insert length (without inspection self) seem to be 1000; the rest is for inaccuracies in output length measuring.
    $log -= 10000; // LOGMAXMARGIN; hardcoded value used elsewhere.
    $get = $absolute; // Admin is free to set a lower value.
    if ($init) {
      variable_set('inspect_outputmax_log', self::$_outputMax_log = $log);
      variable_set('inspect_outputmax_get', self::$_outputMax_get = $get);
      return TRUE;
    }
    return array(
        'logDefault' => $logDefault,
        'log' => $log,
        'get' => $get,
    );
  }

  /**
   * Sets/retrieves dedicated session number hash and request number in either session or cookie.
   *
   * If visitor starts without session, the value will be carried over from cookie to session when session is started,
   * whereas the opposite effectively wont happen (no carry-over from session to cookie).
   *
   * @return void
   */
  protected static function _initSessionNo() {
    if (self::$_sessionNo) {
      return;
    }
    if (drupal_session_started()) {
      $found = FALSE;
      $a =& $_SESSION;
      if (array_key_exists('module', $a)) {
        $a =& $a['module'];
        if (array_key_exists('inspect', $a)) {
          $a =& $a['inspect'];
          if (array_key_exists('sessionNo', $a)) {
            $found = TRUE;
            self::$_sessionNo = $a['sessionNo'];
            self::$_requestNo = (++$a['requestNo']);
          }
        }
        else {
          $a['inspect'] = array();
        }
      }
      else {
        $a['module'] = array('inspect' => array());
      }
      if (!$found) {
        $a =& $_SESSION['module']['inspect'];
        //  If cookie value exists from prior to session start: use that, and delete cookie.
        if (isset($_COOKIE['module_inspect_session'])
            && ($le = strlen($_COOKIE['module_inspect_session']) < 25) // Dont remember how large it may get, but that suffices.
            && preg_match('/^[a-zA-Z\d]+\:\d{1,5}$/', $_COOKIE['module_inspect_session'])
        ) {
          $c = explode(':', $_COOKIE['module_inspect_session']);
          self::$_sessionNo = $a['sessionNo'] = $c[0];
          self::$_requestNo = $a['requestNo'] = ((int)$c[1] + 1);
          setcookie(
              'module_inspect_session', '', time() - 3600,
              $GLOBALS['base_path'], $GLOBALS['cookie_domain'], $GLOBALS['is_https'], TRUE
          );
        }
        else {
          $comps = explode('.', uniqid('', TRUE));
          self::$_sessionNo = $a['sessionNo'] = self::baseConvert($comps[0], 16) . self::baseConvert($comps[1]);
          self::$_requestNo = $a['requestNo'] = 1;
        }
      }
      unset($a);
    }
    else { // Fall back on cookie.
      if (isset($_COOKIE['module_inspect_session'])
          && ($le = strlen($_COOKIE['module_inspect_session']) < 40) // Dont remember how large it may get, but that suffices.
          && preg_match('/^[a-zA-Z\d]+\:\d+$/', $_COOKIE['module_inspect_session'])
      ) {
        $a = explode(':', $_COOKIE['module_inspect_session']);
        self::$_sessionNo = $sess = $a[0];
        self::$_requestNo = $req = ((int)$a[1] + 1);
      }
      else {
        $comps = explode('.', uniqid('', TRUE));
        self::$_sessionNo = $sess = self::baseConvert($comps[0], 16) . self::baseConvert($comps[1]);
        self::$_requestNo = $req = 1;
      }
      @setcookie (
          'module_inspect_session', $sess . ':' . $req, 0,
          $GLOBALS['base_path'], $GLOBALS['cookie_domain'], $GLOBALS['is_https'], TRUE
      );
    }
  }

  /**
   * Secures existance of directory private://module/inspect; optionally subdir to that.
   *
   * @param string $sub_dir
   *  - default: empty
   *  - no starting nor trailing slash
   * @return string|boolean
   *  - FALSE on error
   */
  protected static function _ensureDir($sub_dir = '') {
    if (!($dir = self::$_dir)) {
      if ($dir === NULL) {
        if (!($le = strlen($dir = variable_get('file_private_path', '')))) {
          self::log_db(
              'Private files directory not defined, check /admin/config/media/file-system', 'inspect', WATCHDOG_WARNING
          );
          return (self::$_dir = FALSE);
        }
      }
      else {
        return FALSE;
      }
      if ($dir{$le - 1} == '/') { // Remove trailing slash.
        $dir = substr($dir, 0, $le - 1);
        $le -= 1;
      }
      if ($le < 2) { // Private files dir cannot be shorter than 2 chars (/x).
        return (self::$_dir = FALSE);
      }
      if ($le < 2
          || (!is_dir($dir .= '/module/inspect') && !mkdir($dir, 0755, TRUE))) {
        self::log_db(
            $le < 2 ? 'Private files directory cannot be shorter than 2 chars, check /admin/config/media/file-system' :
                'Directory private://module/inspect is not a dir or cannot be created',
            'inspect',
            WATCHDOG_WARNING
        );
        return (self::$_dir = FALSE);
      }
      self::$_dir = $dir;
    }
    if (!$sub_dir
        || is_dir($dir .= '/' . $sub_dir) || mkdir($dir, 0755, TRUE)) {
      return $dir;
    }
    self::log_db(
        'Directory private://module/inspect/' . $sub_dir . ' is not a dir or cannot be created',
        'inspect',
        WATCHDOG_WARNING
    );
    return FALSE; // Failed to ensure subdir.
  }

  /**
   * @return array|bool
   *  - array: copy of self::$_permissions
   *  - FALSE if no user object, or not permitted to either log or get
   */
  protected static function _init() {
    if (!self::$_initialized) {
      if (!array_key_exists('user', $GLOBALS) || !is_object($GLOBALS['user']) || !property_exists($GLOBALS['user'], 'uid')) {
        return FALSE;
      }
      self::$_userId = $uid = $GLOBALS['user']->uid;
      $isCli = self::$_isCli = drupal_is_cli();
      $log = $get = FALSE;
      if (user_access('inspect log') || $isCli) { // drush always allowed to log.
        $log = self::$_permissions['inspect log'] = TRUE;
      }
      if ( (user_access('inspect get') || $isCli)
          //  ...And isnt admin user 1 (or is user 1 but not prohibited to do get).
          && ($uid != 1 || !variable_get('inspect_disallow_get_admin', FALSE))) {
        $get = self::$_permissions['inspect get'] = TRUE;
      }
      if (!$log && !$get) {
        return FALSE;
      }
      //  Check OS.
      $osNix = self::$_os_nix = strpos(PHP_OS, 'WIN') !== 0 ? 1 : 0;
      //  Find document root, for removal from absolute paths.
      $le0 = strlen($docRoot = dirname($_SERVER['SCRIPT_FILENAME']));
      $le1 = strlen($realPath = substr(dirname(__FILE__), 0, -1 * (1 + strlen(drupal_get_path('module', 'inspect')))));
      self::$_pathLength = ($le0 < $le1 ? $le0 : $le1) - ($osNix ? 0 : 2); // Remove drive, for Windows.
      self::$_paths = array(
          $osNix ? ('/^' . preg_quote($docRoot, '/') . '\//') :
              ('/^(' . $docRoot{0} . '\:)?' . preg_quote(substr($docRoot, 2), '/') . '\//')
      );
      if ($realPath != $docRoot) {
        self::$_paths[] = $osNix ? ('/^' . preg_quote($realPath, '/') . '\//') :
              ('/^(' . $realPath{0} . '\:)?' . str_replace('\\\\', '[\\\\\/]', preg_quote(substr($realPath, 2))) . '[\\\\\/]/');
      }
      //  Change file_line default?
      if (variable_get('inspect_fileline_omit', FALSE)) {
        static::$_options['file_line'] = FALSE;
      }
      //  Set/get session and request number.
      if (!self::$_sessionNo) {
        self::_initSessionNo();
      }
      //  Establish request time start.
      $t = self::requestTimeMilli();
      //  Establish max lengths.
      self::outputMax(TRUE);
      //  Establish max execution time abort.
      self::$_maxExecTimeout = floor($t / 1000)
          + floor(
              (self::$_maxExecTime = ini_get('max_execution_time'))
              * ((self::$_maxExecTimePercent = variable_get('inspect_exectime_percent', 70)) / 100)
          );

      self::$_initialized = TRUE;
    }
    return self::$_permissions; // Copy
  }

  /**
   * Checks permissions, secures that all inspection options exist and are in accord with permissions.
   *
   * Any method calling Inspect::_inspect() must guarantee to call this method at least once before calling _inspect().
   *
   * If the calling method wants to call _inspect() more than once it must guarantee to re-use the return value of this method
   * for the later calls to _inspect().
   *
   * Sets/updates $_SESSION['module']['inspect']['sessionNo'] if session started (only on first call, initializing).
   *
   * @see Inspect::$_options
   * @param string $target
   *   - allowed values: get|log|file
   * @param mixed $options
   *   - see Inspect::$_options for supported options (plus depth and truncate)
   *   - default: NULL
   *   - array|object: list of options
   *   - string: interprets to message
   *   - integer: interprets to suggested depth
   *   - other type: ignored
   * @return array|bool|NULL
   *   - FALSE if the user isnt allowed to get|log inspections
   *   - NULL if arg $target isnt get or log
   */
  protected static function _checkPermissionsAndOptions($target, $options = NULL) {
    if (!self::$_initialized && !self::_init()) {
      return FALSE;
    }
    //  Check general permissions.
    $hideScalars = FALSE;
    $hidePaths = FALSE;
    if ($target === 'get') {
      if (!self::$_permissions['inspect get']) {
        return FALSE;
      }
      $hidePaths = TRUE; // we always hide paths when getting
    }
    elseif ($target === 'log' || $target === 'file') {
      if (!self::$_permissions['inspect log']) {
        return FALSE;
      }
    }
    else {
      return NULL;
    }
    //  Check/load options.
    $isList = $isArr = FALSE;
    $depthOpt = static::DEPTH_DEFAULT;
    $truncOpt = static::TRUNCATE_DEFAULT;
    $optsDef = $target === 'file' || ($target === 'get' && self::$_isCli) ? array_merge(static::$_options, static::$_options_file) : static::$_options;
    switch (gettype($options)) {
      case 'NULL':
        $opts =& $optsDef;
        break;
      case 'array':
        if (count($options)) {
          $isList = $isArr = TRUE;
          $opts = $options;
        }
        else {
          $opts =& $optsDef;
        }
        break;
      case 'object':
        $isList = TRUE;
        $opts = get_object_vars($options);
        break;
      case 'integer':
        $depthOpt = $options;
        $opts =& $optsDef;
        break;
      case 'string':
        $opts =& $optsDef;
        $opts['message'] = $options;
        break;
      default:
        //  Ignore argument.
        $opts =& $optsDef;
    }
    //  If list as argument, check validity of the options.
    if ($isList) {
      //  Make sure string needles and replacers arent falsy.
      if (isset($opts['needles']) && !$opts['needles']) {
        unset($opts['needles']);
      }
      if (isset($opts['replacers']) && !$opts['replacers']) {
        unset($opts['replacers']);
      }
      if (empty($opts['one_lined'])) {
        $oneLined = FALSE;
      }
      elseif ($target === 'get') {
        $oneLined = TRUE;
        $depthOpt = 1;
      }
      else {
        unset($opts['one_lined']);
      }
      //  Read depth, hide scalars, truncate, filter, hide paths, if they exist.
      if (isset($opts['depth'])) {
        $depthOpt = (int)$opts['depth'];
      }
      if (!$hideScalars && isset($opts['hide_scalars']) && $opts['hide_scalars']) {
        $hideScalars = TRUE;
      }
      if (isset($opts['truncate'])) {
        if (($v = (int)$opts['truncate'])) {
          if ($v < 0) {
            $v = 0;
          }
          elseif ($v > static::TRUNCATE_MIN) { // If suggested truncate is more than overall min.
            $v = static::TRUNCATE_MIN;
          }
        }
        $truncOpt = $v;
      }
      if (isset($opts['filter'])) {
        if (is_array($v = $opts['filter'])) {
          if (!count($v)) {
            unset($opts['filter']);
          }
        }
        elseif (strlen('' . $v)) {
          $opts['filter'] = array(0 => '' . $v);
        }
      }
      //  We set the bucket hide_paths no matter what, and it can only override in negative (safer) manner.
      if (!$hidePaths && isset($opts['hide_paths']) && $opts['hide_paths']) {
        $hidePaths = TRUE;
      }
      //  If array, some options may also be set as value (and thus numeric key) ~ interpreted as truthy boolean.
      if ($isArr) {
        $a = $opts; // Iterate copy, to allow modification.
        foreach ($a as $k => $v) {
          if (is_string($v) && ('' . (int)$k) == ('' . $k)) { // ~ Integer key, don't rely on type of PHP array numeric keys.
            switch ($v) {
              case 'hide_scalars':
                $opts['hide_scalars'] = $hideScalars = TRUE;
                unset($opts[$k]);
                break;
              case 'hide_paths':
                $opts['hide_paths'] = $hidePaths = TRUE;
                unset($opts[$k]);
                break;
              case 'by_user':
                if ($target === 'file') {
                  $opts['by_user'] = TRUE;
                }
                unset($opts[$k]);
                break;
              case 'one_lined':
                if ($target === 'get') {
                  $opts['one_lined'] = $oneLined = TRUE;
                }
                unset($opts[$k]);
                break;
              case 'no_folding':
              case 'no_preface':
                $opts[$v] = TRUE;
                unset($opts[$k]);
                break;
            }
          }
        }
      }
      //  load defaults, and make arg options override
      $opts = !$oneLined ? array_merge($optsDef, $opts) : array_merge($optsDef, static::$_options_oneLined, $opts);
    }
    //  Check depth.
    if ($depthOpt < 0) {
      $depthOpt = 0;
    }
    elseif ($depthOpt && $depthOpt > static::DEPTH_MAX) {
      $depthOpt = $depth_max;
    }
    //  Load resolved values of depth and truncate into the options array.
    $opts['depth'] = $depthOpt;
    $opts['hide_scalars'] = $hideScalars;
    $opts['truncate'] = $truncOpt;
    $opts['hide_paths'] = $hidePaths;
    //  Escape name and message options
    if (!empty($opts['name'])) {
      if (strlen($u = $opts['name'] = htmlspecialchars(
          $opts['name'],
          ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
          'UTF-8',
          FALSE // No double encoding
      )) > 255) {
        $opts['name'] = substr($u, 0, 255);
      }
    }
    else {
      $opts['name'] = '';
    }
    if (!empty($opts['message'])) {
      if (strlen($u = $opts['message'] = htmlspecialchars(
          $opts['message'],
          ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
          'UTF-8',
          FALSE // No double encoding
      )) > 255) {
        $opts['message'] = substr($u, 0, 255);
      }
    }
    //  Flag whether getting or logging.
    //$opts['target'] = $target; - not used currently
    return $opts;
  }

  /**
   * @param mixed $options
   * @param string $integer
   *  - default: depth
   *  - what to interpretate $options as, if $options is an integer
   * @param string $string
   *  - default: message
   *  - what to interpretate $options as, if $options is a string
   * @return array
   */
  protected static function _resolveOptions($options, $integer = 'depth', $string = 'message') {
    switch (gettype($options)) {
      case 'array':
        $opts = $options;
        break;
      case 'object':
        return get_object_vars($options);
      case 'integer':
        return array($integer => $options);
      case 'string':
        return array($string => $options);
      default:
        return array();
    }
    //  If array, some options may also be set as value (and thus numeric key) ~ interpreted as truthy boolean.
    $a = $opts; // Iterate copy, to allow modification.
    foreach ($a as $k => $v) {
      if (is_string($v) && ('' . (int)$k) == ('' . $k)) { // ~ Integer key, don't rely on type of PHP array numeric keys.
        switch ($v) {
          case 'hide_scalars':
          case 'hide_paths':
          case 'one_lined':
          case 'no_folding':
          case 'no_preface':
            //  Clean up, make value key.
            unset($opts[$k]);
            $opts[$v] = TRUE;
            break;
        }
      }
    }
    return $opts;
  }

  /**
   * The actual variable analyser - do not call it with an options arg that hasnt been checked by _checkPermissionsAndOptions()
   *
   *  Array/object value of $options (any number of options):
   *   - (integer) depth (default 10, max 10 ~ Inspect::DEPTH_DEFAULT|DEPTH_MAX)
   *   - (integer) truncate (default 1000, max 10000 ~ Inspect::TRUNCATE_MIN|TRUNCATE_DEFAULT)
   *   - (bool) hide_scalars: hide values of scalar vars, and type of resource (also supported as value 'hide_scalars')
   *   - (bool) hide_paths: only relevant for log (paths always hidden for get) (also supported as value 'hide_paths')
   *   - (string) name: '$var_name' or "\$var_name", must be escaped, will be truncated to 255 (default empty, use 'none' to omit)
   *   - (string) message, will be truncated to 255 (default empty)
   *   - (string) category: logging category (default inspect)
   *   - (integer) severity: default WATCHDOG_DEBUG
   *   - (string|array) filter: filter out that|those key name(s) (default empty)
   *   - (array) needles: list of char|string needles to be used in str_replace (default newline carriage return < > " ')
   *   - (array) replacers: list of char|string replacers to be used in str_replace (default equivalent xml entities)
   *   - (boolean) one_lined: create one-lined formatting (getting only, and also supported as value 'one_lined')
   *   - (boolean) no_folding: dont make output expansible/collapsible (also supported as value 'no_folding')
   *   - (boolean) file_line: display file and line of call to inspect function/method (default do)
   *   - (bool) no_preface: truthy to omit [Inspect... preface
   *   - (string) quote_begin: char|string used as initial quote (default >>)
   *   - (string) quote_end: char|string used as final quote (default <<)
   *   - (array) delimiters: between object/array buckets (default newlines)
   *   - (string) indent: indentation of object/array buckets (default 2 spaces)
   *   - (string) enclose_tag: html enclosure tag for the log as a whole (default pre)
   *   - (string) newline: newline character (default newline)
   *
   *   String truncation:
   *   - truncate option less than one (hide strings) doesnt apply to strings that solely consist of digits (stringed integers)
   *   - a string bucket of an object/array keyed pass or password (lowercase) will always be hidden (no matter it's content)
   *
   *   Removes paths at the beginning of variables:
   *   - always removes document root
   *   - always removes relative and absolute paths when getting (optional when logging)
   *   - path removal takes place before escaping and truncating (except for overly long strings)
   *
   *   Reports non-integer numbers:
   *   - double as float
   *   - NaN as NaN and infinite as infinite
   *
   *   Reports string lengths and path removal (path reduced to filename):
   *   - un-truncated string, and when hiding scalar values: (multibyte length|ascii length)
   *   - truncated string: (multibyte length|ascii length|truncation length)
   *   - if path removed: there will be a fourth flag (exclamation mark), if no truncation truncation length will be hyphen
   *
   *   Recursive:
   *   - always stops at 10th (Inspect::DEPTH_MAX) recursion or less (~ the depth option)
   *   - when buckets of an object/array is analyzed, buckets of buckets will be analyzed to one lesser depth
   *   - detects if a bucket of an object references the object itself (marked *RECURSION*)
   *   - detects GLOBALS array's self-referencing GLOBALS bucket (marked *RECURSION*)
   *   - doesnt check identity for arrays (not possible in PHP), nor 'same-ness' (too heavy, and prone to err)
   *
   * @throws Exception
   *  - UNCAUGH, if cumulate inspection output length exceeds maximum
   * @param mixed $var
   * @param array &$options
   *   - by reference, but will not be altered
   * @param integer $_curDepth
   *   - default: zero
   *   - recursion counter, do not call it with anything but zero
   * @return string
   */
  protected static function _inspect($var, &$options, $_curDepth = 0) {
    if (!self::$_initialized) {
      return '';
    }
    //  Execution time check, every 1000th time.
    if (!((++self::$_callNo) % 1000) && time() > self::$_maxExecTimeout) {
      throw new Exception(
          'Inspection aborted: was called after ' . self::$_maxExecTimePercent . '% of PHP max_execution_time['
              . self::$_maxExecTime . '] had passed, using options depth[' . $options['depth']
              . '] and truncate[' . $options['truncate'] . '], try less depth or more truncation.',
          101
      );
    }
    //  Output length check.
    if (($u = self::$_outputLength) && $u > self::$_outputMax) {
      throw new Exception(
          'Inspection aborted: output length exceeds maximum ' . self::$_outputMax . ', using options depth[' . $options['depth']
              . '] and truncate[' . $options['truncate'] . '], try less depth or more truncation.',
          100
      );
    }
    $maxDepth = $options['depth'];
    //  Recursion limit check.
    if ($_curDepth
        && $_curDepth > $maxDepth) {
      throw new Exception('Algo error, current depth ' . $_curDepth . ' exceeds max depth ' . $maxDepth . '.');
    }
    //  Object or array.
    $tArr = FALSE;
    if (is_object($var) || ($tArr = is_array($var))) {
      $s = $tArr ? 'array' : ($tObj = get_class($var));
      if (!$_curDepth && !strlen($options['name'])) {
        $options['name'] = $s;
      }
      $s = '(' . $s . ':';
      //  If at max depth: simply get length of the container.
      if ($_curDepth == $maxDepth) {
        if ($tArr) {
          $s .= ($le = count($var)) . ')' . (!$le ? '' : ' [...]');
        }
        else { // stdClass cannot be counted, simply iterate
          $le = 0;
          foreach ($var as $sub) {
            ++$le;
          }
          $s .= $le . ')' . (!$le ? '' : ' {...}');
        }
        self::$_outputLength += strlen($s);
        return $s;
      }
      //  Dive into container buckets.
      else {
        $delimFirst = $options['delimiters'][0]
            . $options['pre_indent'] . str_repeat($indent = $options['indent'], $_curDepth + 1);
        $delimLater = $options['delimiters'][1]
            . $options['pre_indent'] . str_repeat($indent, $_curDepth + 1);
        $le = 0;
        $sSub = '';
        //  Different iterations for object vs array, because:
        //  - array ust iterate by explicit reference,
        //    whereas object may fail fatally doing that (An iterator cannot be used with foreach by reference)
        //  - they must have different self reference checks
        //  - array must be checked for (all) numeric indices
        if ($tArr) {
          $numArr = TRUE; // if all numeric indices, use [] instead of {}
          self::$_outputLength += strlen(
              $s
          );
          foreach ($var as $key => &$sub) { // by explicit reference
            $delim = (++$le) > 1 ? $delimLater : $delimFirst;
            //  GLOBALS check, skip self reference.
            if ($key === 'GLOBALS' && is_array($sub) && array_key_exists('GLOBALS', $sub)) {
              $numArr = FALSE;
              self::$_outputLength += strlen(
                  $u = $delim . 'GLOBALS: (array) *RECURSION*'
              );
              $sSub .= $u;
              continue;
            }
            if ($numArr && $key !== $le - 1) { // numeric array check
              $numArr = FALSE;
            }
            if ($options['filter'] && in_array($key, $options['filter'], TRUE)) {
              self::$_outputLength += strlen(
                  $u = $delim . $key . ': F'
              );
              $sSub .= $u;
            }
            else {
              $options['key'] = $key; // pass key, for password check
              self::$_outputLength += strlen(
                  $u = $delim . $key . ': '
              );
              $sSub .= $u . self::_inspect($sub, $options, $_curDepth + 1); // recursion
            }
          }
          self::$_outputLength += strlen(
              $u = $le . ') '
                  . ($numArr ? '[' : '{')
          );
          $s .= $u . $sSub;
          self::$_outputLength += strlen(
              $u = $options['delimiters'][2] . str_repeat($indent, $_curDepth)
                  . ($numArr ? ']' : '}')
          );
          return $s . $u;
        }
        else { // object
          self::$_outputLength += strlen(
              $s
          );
          foreach ($var as $key => $sub) { // not &$sub, iterating object is implicitly by reference
            $delim = (++$le) > 1 ? $delimLater : $delimFirst;
            if ($sub === $var) { // check for identity (self reference)
              self::$_outputLength += strlen(
                  $u = $delim . $key . ': (' . $tObj . ') *RECURSION*'
              );
              $sSub .= $u;
              continue;
            }
            if ($options['filter'] && in_array($key, $options['filter'], TRUE)) {
              self::$_outputLength += strlen(
                  $u = $delim . $key . ': F'
              );
              $sSub .= $u;
            }
            else {
              $options['key'] = $key; // pass key, for password check
              self::$_outputLength += strlen(
                  $u = $delim . $key . ': '
              );
              $sSub .= $u . self::_inspect($sub, $options, $_curDepth + 1); // recursion
            }
          }
          self::$_outputLength += strlen(
              $u = $le . ') {'
          );
          $s .= $u . $sSub;
          self::$_outputLength += strlen(
              $u = $options['delimiters'][2] . str_repeat($indent, $_curDepth)
                  . '}'
          );
          return $s . $u;
        }
      }
    }
    //  Scalars, NULL and resource.
    switch (($t = gettype($var))) {
      case 'boolean':
        self::$_outputLength += strlen(
            $s = '(' . $t . ')' . (!$options['hide_scalars'] ? ($var ? ' TRUE' : ' FALSE') : '')
        );
        return $s;
      case 'integer':
      case 'double':
      case 'float':
        if (!$options['hide_scalars'] && !is_finite($var)) {
          self::$_outputLength += strlen(
              $s = '(' . (is_nan($var) ? 'NaN' : 'infinite') . ')'
          );
          return $s;
        }
        self::$_outputLength += strlen(
            $s = '(' . ($t == 'double' ? 'float' : $t) . ')'
                . ( $options['hide_scalars'] ? '' :
                    (!$var ? ' 0' : (' ' . self::numberToString($var)) ) )
        );
        return $s;
      case 'string':
        //  (string:n0|n1|n2): n0 ~ multibyte length, n1 ~ ascii length, n2 only occurs if truncation.
        $s = '(string:';
        //  Get multibyte length and ascii length, and check for invalid UTF-8.
        $leRaw = strlen($var);
        if (!drupal_validate_utf8($var)) {
          self::$_outputLength += strlen(
              $s .= '?|' . $leRaw . '|0) *INVALID_UTF8*'
          );
          return $s;
        }
        $leMb = $leRaw ? self::_mb_strlen($var) : 0;
        $s .= !$leRaw ? '0|0' : ($leMb . '|' . $leRaw);
        //  If empty or hide scalar values.
        if ($options['hide_scalars'] || !$leRaw) {
          self::$_outputLength += strlen(
              $s .= ')'
                  . ( $options['hide_scalars'] ? '' : // no quotes if hide value
                      (' ' . $options['quote_begin'] . $options['quote_end']) )
          );
          return $s;
        }
        //  If very long string, truncate to absolute max before working on the string internals.
        $truncedTo = 0;
        if ($leMb > ($trunc = $options['truncate']) && $leMb > ($trunc_min = static::TRUNCATE_MIN)) {
          $var = self::_mb_truncate($var, $trunc_min);
          $truncedTo = $leMb = $trunc_min;
        }
        //  If truncate is zero.
        if (($pw = ($_curDepth > 0 && ($options['key'] === 'pass' || $options['key'] === 'password'))) || $trunc < 1) {
          //  Look for stringed integer (consists solely of digits), but only if not truncated.
          if (!$pw && !$truncedTo && preg_match('/^\d+$/', $var)) {
            self::$_outputLength += strlen(
                $s .= ') ' . $options['quote_begin'] . $var . $options['quote_end'] // Stringed integer.
            );
            return $s;
          }
          self::$_outputLength += strlen(
              $s .= '|0) ' . $options['quote_begin'] . '...' . $options['quote_end'] // |0 ~ flag truncation
          );
          return $s;
        }
        //  Remove document root (always) and/or path.
        //  Do it before escaping (needles may contain dot or slash).
        $pathRem = FALSE;
        if ($leRaw > 3) {
          if ($options['hide_paths']) { //  This algo also removes document root.
            if (self::$_os_nix) { // Non-windows.
              if ($var{0} === '/' // Absolute path.
                  || ( $var{0} === '.' && (strpos($var, '../') === 0 || strpos($var, './') === 0) ) // Relative.
              ) {
                $pathRem = TRUE;
              }
            }
            else { // Windows.
              if ( strpos($var, '/') !== FALSE || strpos($var, '\\') ) { // Has slash.
                if (($char = $var{0}) === '.') { // Check for relative.
                  if (strpos($var, '../') === 0 || strpos($var, './') === 0
                        || strpos($var, '..\\') === 0 || strpos($var, '.\\') === 0) {
                    $pathRem = TRUE;
                  }
                }
                //  Absolute path (ascii letter + : + slash).
                //elseif (preg_match('/^[A-Za-z]\:[\/\x5C]/', $var)) {
                elseif ( ( ( ($char = ord($char)) >= 65 && $char <= 90) || ($char >= 97 && $char <= 122) )
                    && $var{1} === ':'
                    && (($char = $var{2}) === '/' || $char === '\\')
                ) {
                  $pathRem = TRUE;
                }
              }
            }
            if ($pathRem) {
              $leMb = self::_mb_strlen($var = basename($var));
            }
          }
          elseif ($leRaw >= self::$_pathLength) {
            if (self::$_os_nix) { // non-windows
              if ($var{0} === '/' && strlen($var = preg_replace(self::$_paths, '', $var)) < $leRaw) {
                $pathRem = TRUE;
                $leMb = self::_mb_strlen($var);
                $var = 'document_root/' . $var;
              }
            }
            else {
              if (strlen($var = preg_replace(self::$_paths, '', $var)) < $leRaw) {
                $pathRem = TRUE;
                $leMb = self::_mb_strlen($var);
                $var = 'document_root/' . $var;
              }
            }
          }
        }
        //  Escape string.
        $var = htmlspecialchars(
            str_replace($options['needles'], $options['replacers'], $var),
            ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
            'UTF-8',
            FALSE // No double encoding
        );
        //  Truncate string.
        if ($leMb > $trunc) {
          $var = self::_mb_truncate($var, $trunc);
          $truncedTo = $trunc;
        }
        self::$_outputLength += strlen(
            $s .= (!$truncedTo ? (!$pathRem ? '' : '|-|!' ) :
                    ('|' . $truncedTo . (!$pathRem ? '' : '|!' ) ) )
                . ') '
                . $options['quote_begin'] . (!$pathRem ? '' : '...' ) . $var . (!$truncedTo ? '' : '...') . $options['quote_end']
        );
        return $s;
      case 'resource':
        self::$_outputLength += strlen(
            $s = '(resource)' . (!$options['hide_scalars'] ? (' ' . get_resource_type($var) ) : '')
        );
        return $s;
      default: // NULL and unknown type
        self::$_outputLength += strlen(
            $s = '(' . $t . ')'
        );
        return $s;
    }
  }

  /**
   * Get file and line of outmost call to an inspect function or public method.
   *
   * @param array|NULL $trace
   *  - slightly cheaper when given debug_backtrace() arg
   * @param boolean $hidePaths
   *  - default: TRUE
   * @return string
   */
  protected static function _fileLine($trace = NULL, $hidePaths = TRUE) {
    if ($trace) {
      if (($le = count($trace)) > 3) {
        $le = 3;
      }
    }
    elseif (($le = count($trace = debug_backtrace())) > 4) {
      $le = 4;
    }
    //  finds last frame having an Inspect class or inspect function
    $iFrm = -1;
    for ($i = 0; $i < $le; $i++) {
      $nspct = FALSE;
      if (isset($trace[$i]['class']) && strlen($s = $trace[$i]['class'])) {
        $nspct = (strpos($s, 'Inspect') === 0);
      }
      else {
        $nspct = (isset($trace[$i]['function']) && strlen($s = $trace[$i]['function'])
              && ($s == 'inspect' || strpos($s, 'inspect_') === 0) );
      }
      if ($nspct) {
        $iFrm = $i;
      }
      elseif ($iFrm > -1) { // Break a first non-inspect method/function after an inspect method/function was found.
        break;
      }
    }
    return '@'
        . ($iFrm > -1 ?
            (
                (isset($trace[$iFrm]['file']) && strlen($s = trim($trace[$iFrm]['file'])) ?
                  ($hidePaths ? basename($s) : ('document_root/' . preg_replace(self::$_paths, '', $s)) ) :
                  'unknown'
                )
                . ':'
                . (isset($trace[$iFrm]['line']) ? trim($trace[$iFrm]['line']) : '?')
            ) : 'unknown');
  }

  /**
   * Resolve constants.
   *
   * @param string|object|falsy $cat_or_instance_or_classname
   *   - default: 'user' (~ the user category of defined constants)
   *   - falsy: all defined constants, listed by category
   *   - 'class SomeClassName': class constants of class SomeClassName
   *   - other strings: that category of defined constants
   *   - object: class constants of the object's class
   * @param bool $keySort
   *   - default: falsy (~ do not sort constants, by name)
   * @return array|NULL
   *   - array, buckets: array to be inspected, name
   *   - FALSE on error
   */
  protected static function _constants($cat_or_instance_or_classname = 'user', $keySort = FALSE) {
    $cat = $cat_or_instance_or_classname;
    $nm = 'constants';
    if (!$cat) {
      $a = get_defined_constants(TRUE);
    }
    elseif (is_string($cat)) {
      $cat = preg_replace('/[^a-zA-Z\d_\-\ ]/', '', $cat);
      if (substr($cat, 0, 6) === 'class ') {
        if (($c = substr($cat, 6)) && class_exists($c)) {
          $o = new ReflectionClass($c);
          $a = $o->getConstants();
          $nm = 'class ' . $c . ' ' . $nm;
        }
        else {
          return FALSE;
        }
      }
      else {
        $a = get_defined_constants($cat);
        $a = array_key_exists($cat, $a) ? $a[$cat] : array();
        $nm = $cat . ' ' . $nm;
      }
    }
    elseif (is_object($cat)) {
      $c = get_class($cat);
      $o = new ReflectionClass($c);
      $a = $o->getConstants();
      $nm = 'class ' . $c . ' ' . $nm;
    }
    else {
      return FALSE;
    }
    if ($keySort) {
      ksort($a);
    }
    return array($a, $nm);
  }

  /**
   * Get request time start - in milliseconds - more accurately than the REQUEST_TIME constant.
   *
   * The value is usually slightly higher than REQUEST_TIME, apparantly the latter is a floor'ed value.
   *
   * Uses either $_SERVER['REQUEST_TIME_FLOAT'] (PHP>=5.4), request header X-Request-Received-Processed, or REQUEST_TIME.
   *
   * To enable use of the custom request header (Apache only), add the following to document root .htaccess:
   * @code
   * <IfModule mod_headers.c>
   *   RequestHeader set X-Request-Received-Processing "%t %D"
   * </IfModule>
   * @endcode
   *
   * @return float
   */
  public static function requestTimeMilli() {
    if (!($rtm =& self::$_requestTimeMilli)) {
      $rtm = ($rts = REQUEST_TIME) * 1000;
      if (array_key_exists('REQUEST_TIME_FLOAT', $_SERVER) && ($v = (int) $_SERVER['REQUEST_TIME_FLOAT'])) {
        $rtm = round($v * 1000, 3);
      }
      elseif (
          ( ( array_key_exists('HTTP_X_REQUEST_RECEIVED_PROCESSING', $_SERVER) && ($v = $_SERVER['HTTP_X_REQUEST_RECEIVED_PROCESSING']) )
          || ( function_exists('apache_request_headers') && ($a = apache_request_headers())
              && array_key_exists('X-Request-Received-Processing', $a) && ($v = $a['X-Request-Received-Processing']) ) )
          && strlen($v) < 31 && preg_match('/^t\=\d{16,17}\ D\=\d{1,9}/', $v)
      ) {
        $a = explode(' ', $v);
        $proc = (float) substr($a[1], 2);
        //  If received time can be reduced to positive integer, and that is close to REQUEST_TIME.
        if ( ($iRec = (int) round(($flRec = (float) substr($a[0], 2)) / 1000000)) > 0 && $iRec <= $rts + 1 && $iRec >= $rts - 1) {
          $rtm = round(($flRec + $proc) / 1000, 3);
        }
      }
    }
    return $rtm;
  }

  /**
   * Log to database (currently watchdog).
   *
   * No permissions check for this method - everybody is allowed to log a message.
   *
   * Checks message length against maximum allowed database query length.
   *
   * @param string $message
   *   - default: empty string
   * @param string $category
   *   - default: inspect
   * @param integer $severity
   *   - default: WATCHDOG_DEBUG
   * @return bool
   *  - FALSE: failing to initialize, or on database query error
   */
  public static function log_db($message = '', $category = 'inspect', $severity = WATCHDOG_DEBUG) {
    if (!self::$_initialized && !self::_init()) {
      return FALSE;
    }
    try {
      //  10000 ~ LOGMAXMARGIN; hardcoded value used elsewhere.
      if (strlen($message) > self::$_outputMax_log + 10000) {
        $message = 'Message length [' . strlen($message) . '] exceeds max allowed [' . (self::$_outputMax_log + 10000) . ']';
        $severity = WATCHDOG_WARNING;
      }
      watchdog(
          substr($category, 0, 64), // Will generate error when extracting strings for translation, but cant do anything about.
          '!message', // To prevent error when extracting strings for translation.
          array('!message' => $message), // To prevent error when extracting strings for translation.
          $severity > -1 ? $severity : WATCHDOG_DEBUG
      );
      return TRUE;
    }
    catch (Exception $xc) {
    }
    return FALSE;
  }

  /**
   * Log to file.
   *
   * No permissions check for this method - everybody is allowed to log a message.
   *
   * @param string $message
   *   - default: empty string
   * @param string $category
   *   - default: inspect
   * @param integer $severity
   *   - default: WATCHDOG_DEBUG
   * @param boolean $by_user
   *   - default: FALSE (~ logs to common log file)
   *   - truthy: logs to user specific log file
   * @return bool
   *  - FALSE: failing to initialize, or failing to file
   */
  public static function log_file($message = '', $category = 'inspect', $severity = WATCHDOG_DEBUG, $by_user = FALSE) {
    if ( (!self::$_initialized && !self::_init())
        || (!($dir = self::$_logDir) && !($dir = self::_ensureDir('logs'))) ) {
      return FALSE;
    }
    switch ((int) $severity) {
      case WATCHDOG_DEBUG:
        $sev = 'debug';
        break;
      case WATCHDOG_ALERT:
        $sev = 'alert';
        break;
      case WATCHDOG_CRITICAL:
        $sev = 'critical';
        break;
      case WATCHDOG_ERROR:
        $sev = 'error';
        break;
      case WATCHDOG_WARNING:
        $sev = 'warning';
        break;
      case WATCHDOG_NOTICE:
        $sev = 'notice';
        break;
      case WATCHDOG_INFO:
      case -1:
        $sev = 'info';
        break;
      default: // 0
        $sev = 'emergency';
    }
    return file_put_contents(
        $dir . '/inspect_' . date('Ymd', (int) round(self::$_requestTimeMilli / 1000))
            . (!$by_user ? '' : // For anonymous user: avoid filing concurrency by using the session number.
                ('_user_' . self::$_userId . (self::$_userId ? '' : ('_' . self::$_sessionNo)))
            )
            . '.log',
        self::$_sessionNo . ':' . self::$_requestNo
            . ' - ' . date('Y-m-d H:i:s', floor(self::$_requestTimeMilli / 1000)) . ' - user: ' . self::$_userId
            . ' --------------------------------------------------'
            . "\n" . $_SERVER['REQUEST_METHOD'] . ' ' . request_uri()
            . "\nseverity+category: " . $sev . ' - ' . $category
            . "\n" . $message . "\n\n",
        FILE_APPEND
    ) ? TRUE : FALSE;
  }

  /**
   * Inspect variable and log to database (watchdog), if permitted.
   *
   *   Like PHP native var_dump()/var_export() except:
   *   - is controlled by permissions
   *   - has much more options, and far better formatting (JSONesque)
   *   - doesnt fail on inspecting $GLOBALS
   *   - aborts if producing too large output, or taking too much time
   *   - doesnt echo, logs to database log
   *
   *   Integer value of $options (depth):
   *   - the value controls how deeply the variable is going to be analyzed
   *   - default: 10 (Inspect::DEPTH_DEFAULT)
   *   - max: 10 (Inspect::DEPTH_MAX)
   *
   *   Array/object value of $options (any number of options):
   *   - (integer) depth (default 10, max 10 ~ Inspect::DEPTH_DEFAULT|DEPTH_MAX)
   *   - (integer) truncate (default 1000, max 10000 ~ Inspect::TRUNCATE_MIN|TRUNCATE_DEFAULT)
   *   - (boolean) hide_scalars: hide values of scalar vars, and type of resource (also supported as value 'hide_scalars')
   *   - (boolean) hide_paths: only relevant for log (paths always hidden for get) (also supported as value 'hide_paths')
   *   - (string) name: '$var_name' or "\$var_name", must be escaped, will be truncated to 255 (default empty, use 'none' to omit)
   *   - (string) message, will be truncated to 255 (default empty)
   *   - (string) category: logging category (default inspect)
   *   - (integer) severity: default WATCHDOG_DEBUG
   *   - (string|array) filter: filter out that|those key name(s) (default empty)
   *   - (array) needles: list of char|string needles to be used in str_replace (default newline carriage return < > " ')
   *   - (array) replacers: list of char|string replacers to be used in str_replace (default equivalent xml entities)
   *   - (boolean) one_lined: create one-lined formatting (getting only, and also supported as value 'one_lined')
   *   - (boolean) no_folding: dont make output expansible/collapsible (also supported as value 'no_folding')
   *   - (boolean) file_line: display file and line of call to inspect function/method (default do)
   *
   *   String truncation:
   *   - truncate option less than one (hide strings) doesnt apply to strings that solely consist of digits (stringed integers)
   *   - a string bucket of an object/array keyed pass or password (lowercase) will always be hidden (no matter it's content)
   *
   *   Removes paths at the beginning of variables:
   *   - always removes document root
   *   - always removes relative and absolute paths when getting (optional when logging)
   *
   *   Reports non-integer numbers:
   *   - double as float
   *   - NaN as NaN and infinite as infinite
   *
   *   Reports string lengths and path removal (path reduced to filename):
   *   - un-truncated string, and when hiding scalar values: (multibyte length|ascii length)
   *   - truncated string: (multibyte length|ascii length|truncation length)
   *   - if path removed: there will be a fourth flag (exclamation mark), if no truncation truncation length will be hyphen
   *
   *   Recursive:
   *   - detects if a bucket of an object references the object itself (marked *RECURSION*)
   *   - detects GLOBALS array's self-referencing GLOBALS bucket (marked *RECURSION*)
   *   - doesnt check identity for arrays (not possible in PHP), nor 'same-ness' (too heavy, and prone to err)
   *
   * Executes variable analysis within try-catch.
   *
   * @param mixed $var
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to suggested depth
   *   - string: interprets to message
   *   - other type: ignored
   * @return bool|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   */
  public static function log($var, $options = NULL) {
    if (!($opts = self::_checkPermissionsAndOptions('log', $options)) ) {
      return NULL;
    }
    $nl = $opts['newline'];
    try {
      self::$_outputMax = self::$_outputMax_log;
      self::$_outputLength = 0;
      $nspct = self::_inspect($var, $opts);
    }
    catch (Exception $xc) {
      switch ((int)$xc->getCode()) {
        case 100: // Output length exceeding maximum.
          $nspct = $xc->getMessage();
          $opts['severity'] = WATCHDOG_WARNING;
          //  Try again?
          $t = $opts['truncate'];
          if (($d = $opts['depth']) > 2 || $t > 100) {
            $opts['depth'] = $d > 2 ? ceil($d / 2) : $d;
            $opts['truncate'] = $t > 100 ? ceil($t / 2) : $t;
            $nspct .= $nl . 'Trying again, using options depth[' . $opts['depth'] . '] and truncate[' . $opts['truncate'] . '].';
            try {
              self::$_outputLength = 0;
              $nspct .= $nl . self::_inspect($var, $opts);
            }
            catch (Exception $xc1) {
              if ($xc->getCode() != 100) { // 100 ~ Output length exceeding maximum.
                return '';
              }
              $nspct .= $nl . $xc1->getMessage();
            }
          }
          break;
        case 101: // Max percent of max_execution_time passed.
          $nspct = $xc->getMessage();
          $opts['severity'] = WATCHDOG_WARNING;
          break;
        default:
          return FALSE;
      }
    }
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag'])) {
      $tagStart = '<' . $u . '  class="module-inspect' . (empty($opts['no_folding']) ? '-collapsible' : '') . '">';
      $tagEnd = '</' . $u . '>';
    }
    return self::log_db(
        $tagStart
          . ( empty($opts['message']) ? '' : ($opts['message'] . ':' . $nl) )
          . ( empty($opts['no_preface']) ? ('[Inspect - ' . self::$_sessionNo . ':' . self::$_requestNo . ':' . (++self::$_logNo)
              . ' - depth:' . $opts['depth'] . ']') : '')
          . ( empty($opts['name']) || $opts['name'] == 'none' ? '' : (' ' . $opts['name']) ) . $nl
          . ( empty($opts['file_line']) ? '' : (self::_fileLine(debug_backtrace(), $opts['hide_paths']) . $nl) )
          . $nspct
          . $tagEnd,
        $opts['category'],
        array_key_exists('severity', $opts) ? $opts['severity'] : -1
    );
  }

  /**
   * Inspect variable and log to file, if permitted.
   *
   *   Array/object value of $options are like Inspect::log()'s options, except:
   *   - (boolean) by_user: log to user specific file instead of common file
   *
   * @param mixed $var
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to suggested depth
   *   - string: interprets to message
   *   - other type: ignored
   * @return bool|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   */
  public static function file($var, $options = NULL) {
    if (!($opts = self::_checkPermissionsAndOptions('file', $options)) ) {
      return NULL;
    }
    $nl = "\n";
    try {
      self::$_outputMax = self::$_outputMax_get; // Use getting output max.
      self::$_outputLength = 0;
      $nspct = self::_inspect($var, $opts);
    }
    catch (Exception $xc) {
      switch ((int)$xc->getCode()) {
        case 100: // Output length exceeding maximum.
          $nspct = $xc->getMessage();
          $opts['severity'] = WATCHDOG_WARNING;
          //  Try again?
          $t = $opts['truncate'];
          if (($d = $opts['depth']) > 2 || $t > 100) {
            $opts['depth'] = $d > 2 ? ceil($d / 2) : $d;
            $opts['truncate'] = $t > 100 ? ceil($t / 2) : $t;
            $nspct .= $nl . 'Trying again, using options depth[' . $opts['depth'] . '] and truncate[' . $opts['truncate'] . '].';
            try {
              self::$_outputLength = 0;
              $nspct .= $nl . self::_inspect($var, $opts);
            }
            catch (Exception $xc1) {
              if ($xc->getCode() != 100) { // 100 ~ Output length exceeding maximum.
                return '';
              }
              $nspct .= $nl . $xc1->getMessage();
            }
          }
          break;
        case 101: // Max percent of max_execution_time passed.
          $nspct = $xc->getMessage();
          $opts['severity'] = WATCHDOG_WARNING;
          break;
        default:
          return FALSE;
      }
    }
    return self::log_file(
        ( empty($opts['message']) ? '' : ($opts['message'] . ':' . $nl) )
          . ( empty($opts['no_preface']) ? ('[Inspect - ' . self::$_sessionNo . ':' . self::$_requestNo . ':' . (++self::$_logNo)
              . ' - depth:' . $opts['depth'] . ']') : '')
          . ( empty($opts['name']) || $opts['name'] == 'none' ? '' : (' ' . $opts['name']) ) . $nl
          . ( empty($opts['file_line']) ? '' : (self::_fileLine(debug_backtrace(), $opts['hide_paths']) . $nl) )
          . $nspct,
        $opts['category'],
        array_key_exists('severity', $opts) ? $opts['severity'] : -1,
        empty($opts['by_user']) ? FALSE : $opts['by_user']
    );
  }

  /**
   * Inspect variable and get the output as string, if permitted.
   *
   * @param mixed $var
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to suggested depth
   *   - string: interprets to message
   *   - other type: ignored
   * @return string
   *   - empty if user isnt permitted to get inspections, or on error (other error than exceeding max. output length)
   */
  public static function get($var, $options = NULL) {
    if (!($opts = self::_checkPermissionsAndOptions('get', $options)) ) {
      return '';
    }
    $nl = $opts['newline'];
    try {
      self::$_outputMax = self::$_outputMax_get;
      self::$_outputLength = 0;
      $nspct = self::_inspect($var, $opts);
    }
    catch (Exception $xc) {
      switch ((int)$xc->getCode()) {
        case 100: // Output length exceeding maximum.
          $nspct = $xc->getMessage();
          //  Try again?
          $t = $opts['truncate'];
          if (($d = $opts['depth']) > 2 || $t > 100) {
            $opts['depth'] = $d > 2 ? ceil($d / 2) : $d;
            $opts['truncate'] = $t > 100 ? ceil($t / 2) : $t;
            $nspct .= $nl . 'Trying again, using options depth[' . $opts['depth'] . '] and truncate[' . $opts['truncate'] . '].';
            try {
              self::$_outputLength = 0;
              $nspct .= $nl . self::_inspect($var, $opts);
            }
            catch (Exception $xc1) {
              if ($xc->getCode() != 100) { // 100 ~ Output length exceeding maximum.
                return '';
              }
              $nspct .= $nl . $xc1->getMessage();
            }
          }
          break;
        case 101: // Max percent of max_execution_time passed.
          $nspct = $xc->getMessage();
          break;
        default:
          return FALSE;
      }
    }
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag'])) {
      $tagStart = '<' . $u . '  class="module-inspect' . (empty($opts['no_folding']) ? '-collapsible' : '') . '">';
      $tagEnd = '</' . $u . '>';
    }
    return $tagStart
        . ( empty($opts['message']) ? '' : ($opts['message'] . ':' . $nl) )
        . ( empty($opts['no_preface']) ? ('[Inspect - depth:' . $opts['depth'] . ']') : '')
        . ( empty($opts['name']) || $opts['name'] == 'none' ? '' : (' ' . $opts['name']) ) . $nl
        . ( empty($opts['file_line']) ? '' : (self::_fileLine(debug_backtrace(), TRUE) . $nl) )
        . $nspct . $tagEnd;
  }

  /**
   * Inspect arguments of surrounding function/method and log to database (watchdog), if permitted.
   *
   * $options like Inspect::log().
   *
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to suggested depth
   *   - string: interprets to message
   *   - other type: ignored
   * @return bool|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   */
  public static function args($options = NULL, $_get = FALSE) {
    $trc = debug_backtrace();
    $frame = NULL;
    for ($i = 0; $i < 4; $i++) {
      if (isset($trc[$i]['class']) && strlen($s = $trc[$i]['class'])) {
        if (strpos($s, 'Inspect') === 0) {
          continue;
        }
      }
      elseif (isset($trc[$i]['function']) && strlen($s = $trc[$i]['function'])
            && ($s == 'inspect' || strpos($s, 'inspect_') === 0)) {
        continue;
      }
      if (isset($trc[$i]['args'])) {
        $frame = $trc[$i];
        break;
      }
    }
    if ($frame) {
      if ($options === 'one_lined') {
        $options = array('one_lined' => TRUE);
      }
      //  Set options category and name, if empty.
      $opts = self::_resolveOptions($options);
      if (empty($opts['category'])) {
        $opts['category'] = 'inspect args';
      }
      if (empty($opts['name'])) {
        $opts['name'] = (empty($frame['class']) ? '' : ($frame['class'] . (!empty($frame['type']) ? $frame['type'] : '::') ) )
              . (!empty($frame['function']) ? ($frame['function'] . '()') : '' );
      }
      if (!$_get) {
        return self::log($frame['args'], $opts);
      }
      //  If get is one_lined, we want to get rid of the 'array'
      if (($s = self::get($frame['args'], $opts))
          && !empty($opts['one_lined'])) {
        return preg_replace('/(\(\)\s)?\(array\:/', '(', $s, 1);
      }
      return $s;
    }
    return !$_get ? FALSE : '';
  }

  /**
   * Inspect arguments of surrounding function/method and get the output as string, if permitted.
   *
   * Usable for error messages.
   *
   * $options like Inspect::get().
   *
   * @param mixed $var
   * @param mixed $options
   *   - default: 'one_lined' (~ create short single lined format usable for error messages, using depth:1)
   *   - array|object: list of options
   *   - integer: interprets to suggested depth
   *   - string: interprets to message (except for 'one_lined')
   *   - other type: ignored
   * @return string
   *   - empty if user isnt permitted to get inspections, or on error (other error than exceeding max. output length)
   */
  public static function args_get($options = 'one_lined') {
    return self::args($options, TRUE);
  }

  /**
   * Inspect constants and log to database (watchdog), if permitted.
   *
   *   $options like Inspect::log(), except:
   *   - (bool) sort: if truthy, sorts constants by name (except if inspecting all defined constants, then only category is sorted)
   *
   * @see Inspect::log()
   * @param string|object|falsy $cat_or_instance_or_classname
   *   - default: 'user' (~ the user category of defined constants)
   *   - falsy: all defined constants, listed by category
   *   - 'class SomeClassName': class constants of class SomeClassName
   *   - other strings: that category of defined constants
   *   - object: class constants of that object's class
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - string: interprets to message
   *   - boolean: sort constants
   *   - other type: ignored
   * @return bool|NULL|string
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   *   - string if truthy arg $_get
   */
  public static function constants($cat_or_instance_or_classname = 'user', $options = NULL, $_get = FALSE) {
    $sort = FALSE;
    if (($t = gettype($options)) == 'object') {
      $opts = get_object_vars($options);
      $t = 'array';
    }
    else {
      $opts = $options;
    }
    switch ($t) {
      case 'array':
        if (!empty($opts['sort'])) {
          $sort = TRUE;
          unset($opts['sort']);
        }
        break;
      case 'string':
        $opts = array('message' => $opts);
        break;
      case 'boolean':
        $sort = $opts;
        $opts = array();
        break;
      default:
        $opts = array();
    }
    if (!($a = self::_constants($cat_or_instance_or_classname, $sort))) {
      return !$_get ? FALSE : '';
    }
    $opts['name'] = $a[1];
    return !$_get ? self::log($a[0], $opts) : self::get($a[0], $opts);
  }

  /**
   * Inspect constants and get the output as string, if permitted.
   *
   * @see Inspect::get()
   * @param string|object|falsy $cat_or_instance_or_classname
   *   - default: 'user' (~ the user category of defined constants)
   *   - falsy: all defined constants, listed by category
   *   - 'class SomeClassName': class constants of class SomeClassName
   *   - other strings: that category of defined constants
   *   - object: class constants of that object's class
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - string: interprets to message
   *   - boolean: sort constants
   *   - other type: ignored
   * @return string
   *   - empty if user isnt permitted to get inspections, or on error (other error than exceeding max. output length)
   */
  public static function constants_get($cat_or_instance_or_classname = 'user', $options = NULL) {
    return self::constants($cat_or_instance_or_classname, $options, TRUE);
  }

  //  Utility methods.
  /**
   * Becomes the value of php:ini:precision.
   *
   * @var integer $_precision
   */
  protected static $_precision = 0;

  /**
   * Convert number to string avoiding E-notation for numbers outside system precision range.
   *
   * @param float|integer $num
   * @return string
   */
  public static function numberToString($num) {
    if (!($prec = self::$_precision)) {
      $prec = self::$_precision = ini_get('precision');
    }
    return ($num < pow(10, $prec) && $num > pow(10, -$prec)) ?  // if within system precision, just string it
        ('' . $num) : number_format($num, 0, '.', '');
  }

  /**
   * Convert number or hex string to any base through 2 and 62  (~ larger than 16).
   *
   * Base 62 produces up to 35% shorter string than base 16.
   *
   *  Like native base_convert() precision fails for numbers (somewhere) larger than (tested on windows box with ini precision 14):
   *  - base 10: 9007199254740991 (> 15 digits)
   *  - base 16: 1fffffffffffff (> 14 digits)
   *
   * @param integer|float|string $num
   * @param integer $fromBase
   *  - default: 10
   *  - only supports 10 and 16 (if other, returns arg $num unaltered)
   * @param integer $toBase
   *  - default (and max): 62
   * @return string
   */
  public static function baseConvert($num, $fromBase = 10, $toBase = 62) {
    $table = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
    $zero = $table{0};
    $remainder = $num;
    if ($num < 1) {
      if (!$num) {
        return '' . $zero; // zero
      }
      $remainder *= -1;
    }
    if ($toBase == $fromBase) {
      return $remainder;
    }
    switch ($fromBase) {
      case 10:
        break;
      case 16:
        $remainder = base_convert($remainder, 16, 10);
        break;
      default:
        return $remainder;
    }
    if ($toBase > 62) {
      $toBase = 62;
    }
    elseif ($toBase < 2) {
      $toBase = 2;
    }
    $res = '';
    //  Find power ~ required number of digits.
    $pow = 0;
    $divisor = 1;
    $limit = 100; // probably way too high, havent seen more than 10 iterations (1fffffffffffff, with precision 14)
    while ($remainder >= ($test = pow($toBase, ++$pow)) && (--$limit)) {
      $divisor = $test;
      --$limit;
    }
    --$pow;
    //  Do modulo procedure.
    $limit = 100; // probably way too high, havent seen more than 14 iterations (1fffffffffffff, with precision 14)
    while ($remainder > 0 && (--$limit)) {
      if ($pow) {
        if ($divisor == $remainder) {
          return $res . $table{1} . str_repeat($zero, $pow);
        }
        elseif ($divisor > $remainder) {
          $res .= $zero;
        }
        else {
          if ($remainder > $divisor) {
            $res .= $table{ floor($remainder / $divisor) };
            $remainder = fmod($remainder, $divisor);
          }
          else {
            $res .= $table{ $pow };
            $remainder = fmod($remainder, $divisor);
          }
          if ($remainder < 1 && !floor($remainder)) {
            return $res . str_repeat($zero, $pow);
          }
        }
      }
      else {
        return $res . $table{ $remainder };
      }
      $divisor = pow($toBase, --$pow);
    }
    return $res;
  }

  /**
   * Multibyte-safe string length counter which doesnt require the mb_string PHP extension.
   *
   * @param string $str
   * @return integer
   */
  protected static function _mb_strlen($str) {
    static $mb = -1;
    if (!($le = strlen($str))) {
      return 0;
    }
    if ($mb < 0) {
      if (($mb = self::$_mbString_exists) < 0) {
        $mb = self::$_mbString_exists = extension_loaded('mbstring') ? 1 : 0;
      }
    }
    if ($mb) {
      return mb_strlen($str);
    }
    $n = 0;
    $leading = FALSE;
    for ($i = 0; $i < $le; $i++) {
      if (($ord = ord($str{$i})) < 128) {
        ++$n;
        $leading = FALSE;
      }
      elseif ($ord < 192) { // continuation char
        $leading = FALSE;
      }
      else { // leading char
        if (!$leading) { // a sequence of leadings only counts as a single
          ++$n;
        }
        $leading = TRUE;
      }
    }
    return $n;
  }

  /**
   * Multibyte-safe string truncator which doesnt require the mb_string PHP extension.
   *
   * @param string $str
   * @param integer $length
   *  - default: zero
   * @return integer
   */
  protected static function _mb_truncate($str, $length = 0) {
    static $mb = -1;
    if ($length < 1 || !($le = strlen($str))) {
      return '';
    }
    if ($mb < 0) {
      if (($mb = self::$_mbString_exists) < 0) {
        $mb = self::$_mbString_exists = extension_loaded('mbstring') ? 1 : 0;
      }
    }
    if ($mb) {
      return mb_substr($str, 0, $length);
    }
    $n = 0;
    $leading = FALSE;
    for ($i = 0; $i < $le; $i++) {
      if (($ord = ord($str{$i})) < 128) {
        if ((++$n) > $length) {
          return substr($str, 0, $i);
        }
        $leading = FALSE;
      }
      elseif ($ord < 192) { // continuation char
        $leading = FALSE;
      }
      else { // leading char
        if (!$leading) { // a sequence of leadings only counts as a single
          if ((++$n) > $length) {
            return substr($str, 0, $i);
          }
        }
        $leading = TRUE;
      }
    }
    return $str;
  }
}