<?php
/**
 * @file
 * Drupal Inspect module stack trace inspection class
 */

/**
 * Stack trace inspector
 * @static
 * @singleton
 * @category Drupal
 * @package Developer
 * @subpackage Inspect */
class InspectTrace extends Inspect {
  /**
   * Prevents erratic instantiation of this entirely static class
   */
  private function  __construct() {}

  /**
   * Default sub var recursion depth.
   *
   * Overrides Inspect::DEPTH_DEFAULT - when methods of this class uses methods of Inspect (using static::DEPTH_DEFAULT).
   *
   * @type integer
   */
  const DEPTH_DEFAULT = 1;

  /**
   * Minimum string truncation.
   *
   * Overrides Inspect::TRUNCATE_MIN - when methods of this class uses methods of Inspect (using static::TRUNCATE_MIN).
   *
   * @type integer
   */
  const TRUNCATE_MIN = 1000;

  /**
   * Overrides Inspect::TRUNCATE_DEFAULT - when methods of this class uses methods of Inspect (using static::TRUNCATE_DEFAULT).
   *
   * @type integer
   */
  const TRUNCATE_DEFAULT = 100;

  /**
   * Default maximum stack frame depth
   *
   * @type integer
   */
  const TRACE_LIMIT_DEFAULT = 10;

  /**
   * Absolute maximum stack frame depth
   *
   * @type integer
   */
  const TRACE_LIMIT_MAX = 100;

  /**
   * Default trace options.
   *
   * Adds to/overrides options of Inspect::$_options
   *
   *   Options:
   *   - (string) category: inspect trace
   *   - (string) trace_spacer: spacer between frames
   *   - (string) pre_indent: indentation of object/array buckets
   *   - (array) delimiters: see Inspect
   *
   * @var array $_optionsTrace
   */
  protected static $_optionsTrace = array(
    'category' => 'inspect trace',
    'trace_spacer' => '- - - - - - - - - - - - - - - - - - - - - - - - -',
    'pre_indent' => '  ',
    'delimiters' => array(
        "\n", // Before first bucket.
        "\n", // Before later bucket.
        "\n  "  // Before last bucket.
    ),
  );

  /**
   * Analyzes exception and creates derivate as formatted string.
   *
   * @param Exception|falsy $xc
   *   - if falsy: create new trace
   * @param string $target
   *   - allowed values: get|log|file
   * @param mixed $options
   *   - array|object: list of options
   *   - string: interprets to message
   *   - integer: interprets to limit (maximum stack frame depth), not depth
   *   - other type: ignored
   * @return array|bool|NULL
   *   - array: [ (array) options, (string) trace ]
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   */
  protected static function _trc($xc, $target, $options) {
    //  Populate options with default trace options, before passing to parent method.
    $limit = $limitDef = static::TRACE_LIMIT_DEFAULT;
    $limitMax = static::TRACE_LIMIT_MAX;
    $optsDef =& static::$_optionsTrace;
    if (($t = gettype($options)) == 'object') {
      $opts = get_object_vars($options);
      $t = 'array';
    }
    else {
      $opts = $options;
    }
    switch ($t) {
      case 'array':
        if (isset($opts['limit'])) {
          $limit = (int)$opts['limit'];
          unset($opts['limit']);
        }
        $opts = array_merge($optsDef, $opts); // secure $_optionsTrace buckets
        break;
      case 'integer':
        $limit = $opts;
        $opts = $optsDef;
        break;
      case 'string':
        $opts = $optsDef;
        $opts['message'] = $options;
        break;
      default:
        $opts = $optsDef;
    }
    if (!($opts = static::_checkPermissionsAndOptions($target, $opts))) {
      return $opts;
    }
    if ($limit < 1 || $limit > $limitMax) {
      $limit = $limitMax;
    }
    $opts['limit'] = $limit;
    if (!isset($opts['severity'])) {
      $opts['severity'] = WATCHDOG_DEBUG;
    }
    try {
      //  Create trace, if none given by arg.
      if ($xc) {
        if (is_object($xc) && (($u = get_class($xc)) === 'Exception' || is_subclass_of($xc, 'Exception'))) {
          $xcClass = $u;
          $trace = $xc->getTrace();
          if (count($trace) > $limit) {
            array_splice($trace, $limit);
          }
        }
        else {
          throw new Exception('arg $exception type[' . (!is_object($xc) ? gettype($xc) : get_class($xc)) . '] is not Exception or falsy');
        }
      }
      else {
        $xcClass = '';
        $trace = debug_backtrace();
        //  Remove top levels on synthetic trace, first of all the InspectTrace::__construct frame.
        array_shift($trace);
        //  If called via function instead of class method: remove class method call.
        if (isset($trace[0]['class']) && strtolower($trace[0]['class']) === 'InspectTrace'
            && isset($trace[1]['function']) && substr($trace[1]['function'], 0, 13) === 'inspect_trace') {
          array_shift($trace);
        }
        //  Enforce the trace limit.
        if (count($trace) > $limit) {
          array_splice($trace, $limit + 1); // plus one because we need the bucket holding the initial event
        }
      }

      //  Secure output max.
      self::$_outputLength = 0;
      self::$_outputMax = $target == 'log' ? self::$_outputMax_log : self::$_outputMax_get; // Yes, file uses get max.

      //  Format as string.
      $hidePaths = $opts['hide_paths'];
      $delim = $opts['delimiters'][1];
      $preIndent = $opts['pre_indent'];
      $frameSpacer = $opts['trace_spacer'];
      $sFunc = '';
      //  If exception: resolve its origin.
      if ($xcClass) {
        $sTrc = 'Exception (' . $xcClass . ') - code: ' . intval($xc->getCode()) . $delim;
        $file = trim($xc->getFile());
        $line = (($u = $xc->getLine()) ? trim($u) : '?');
        $message = 'message: ' . ($u = $xc->getMessage()) . $delim;
        //  If severe: pass exception message to overall message, if that is empty.
        if ($opts['severity'] != WATCHDOG_DEBUG && empty($opts['message'])) {
          $opts['message'] = $u;
          if (strlen($u = htmlspecialchars(
              $opts['message'],
              ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
              'UTF-8',
              FALSE // No double encoding
          )) > 255) {
            $u = substr($u, 0, 255);
          }
          $opts['message'] = $u;
        }
      }
      else {
        $sTrc = 'Backtrace' . $delim;
        //  Find inspect method/function, which commanded creation of synthetic trace.
        $frame = array_shift($trace);
        if (isset($frame['class'])) {
          $sFunc = 'static method: ' . $frame['class'] . '::' . (isset($frame['function']) ? ($frame['function'] . '()') : '') . $delim;
        }
        elseif (isset($frame['function'])) {
          $sFunc = 'function: ' . $frame['function'] . '()' . $delim;
        }
        if (isset($frame['file'])) {
          $file = trim($frame['file']);
        }
        $line = (isset($frame['line']) ? trim($frame['line']) : '?');
        $message = '';
      }
      $file = ($le = strlen($file)) ? (
          $hidePaths ? basename($file) : (
              (strlen($noRoot = preg_replace(self::$_paths, '', $file)) < $le ? ('document_root/' . $noRoot) : $file)
          )
      ) : 'unknown';
      $sTrc .= 'file: ' . $file . $delim
            . 'line: ' . $line . $delim
            . $sFunc
            . $message;

      //  Iterate stack frames.
      $nFrame = -1;
      foreach ($trace as &$frame) {
        $sTrc .= (++$nFrame) . ' ' . $frameSpacer . $delim;
        //  File and line.
        $file = isset($frame['file']) && ($le = strlen($u = trim($frame['file']))) ? (
            $hidePaths ? basename($u) : (
                (strlen($noRoot = preg_replace(self::$_paths, '', $u)) < $le ? ('document_root/' . $noRoot) : $u)
            )
        ) : 'unknown';
        $sTrc .= 'file: ' . $file . $delim . 'line: ' . (isset($frame['line']) ? trim($frame['line']) : '?') . $delim;
        //  Class, object, function, type.
        $sFunc = isset($frame['function']) && strlen($u = $frame['function']) ? $u : '';
        $sCls = isset($frame['class']) && strlen($u = $frame['class']) ? $u : '';
        $sType = isset($frame['type']) ? trim($frame['type']) : '';
        $sObj = '';
        if (isset($frame['object'])) {
          if (($o = $frame['object'])) {
            //  We dont know if class bucket is present when object is, but using class is cheaper.
            $sObj = $sCls ? $sCls : get_class($o);
            if ($sFunc) {
              $sFunc = 'method: (' . $sObj . ')' . ($sType ? $sType : '->') . $sFunc;
            }
            else { // Is this possible at all? Simply object, no method call??
              $sObj = 'object (' . $sObj . ')';
            }
          }
        }
        elseif ($sFunc) {
          $sFunc = !$sCls ? ('function: ' . $sFunc) : ('static method: ' . $sCls . ($sType ? $sType : '::') . $sFunc);
          //  $frame['object'] doesnt exist.
        }
        //  else {
        //    $frame['object'] doesnt exist
        //    $frame['function'] may exist, containing non-empty string
        //  }
        $sTrc .= ($sFunc ? $sFunc : $sObj) . $delim;
        //  Args.
        if (isset($frame['args'])) {
          $le = count($args = $frame['args']);
          $sArgs = '';
          for ($i = 0; $i < $le; $i++) {
            $sArgs .= $delim . $preIndent
                . self::_inspect(
                    $args[$i],
                    $opts,
                    0 // recursion counter
                  );
          }
          $sTrc .= 'args (' . $le . ')'
              . (!$le ? '' : (': ' . $sArgs) )
              . $delim;
        }
      }
      $sTrc .= 'END ' . $frameSpacer;
      if (($le = strlen($sTrc)) > self::$_outputMax) {
        throw new Exception(
            'Inspection aborted: output length ' . $le . ' exceeds maximum ' . self::$_outputMax
                . ', using options limit[' . $limit . '], depth[' . $opts['depth'] . '] and truncate[' . $opts['truncate'] . ']'
                . ', try smaller limit, less depth or more truncation.',
            100
        );
      }
      return array($opts, $sTrc);
    }
    catch (Exception $xc1) {
      $opts['severity'] = WATCHDOG_WARNING;
      return array($opts, $xc1->getMessage());
    }
  }

  /**
   * Inspect and log stack trace to database (watchdog), if permitted.
   *
   *   Integer value of $options (limit):
   *   - maximum stack frame depth
   *   - default: 10 (InspectTrace::TRACE_LIMIT_DEFAULT)
   *   - max: 100 (InspectTrace::TRACE_LIMIT_MAX)
   *
   *   Array/object value of $options are like Inspect::log()'s options, except:
   *   - (integer) limit (default 10, max 100 ~ InspectTrace::TRACE_LIMIT_DEFAULT|TRACE_LIMIT_MAX)
   *   - (string) trace_spacer: spacer between frames (default hyphen dotted line, length 49)
   *   - (string) category: inspect trace
   *   - (boolean) file_line: is not used by tracer
   *
   * Default variable analysis depth (1 ~ InspectTrace::DEPTH_DEFAULT) is less than for plain variable inspection.
   *
   * String truncation minimum (1000 ~ InspectTrace::TRUNCATE_MIN) is less than for plain variable inspection.
   *
   * See inspect() for specifics of variable analysis.
   *
   * Default watchdog severity: WATCHDOG_DEBUG.
   *
   * Executes variable analysis within try-catch.
   *
   * @see Inspect::log()
   * @param Exception|falsy $exception
   *   - default: NULL (~ create new backtrace)
   *   - if Exception, trace that
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to limit (maximum stack frame depth)
   *   - string: interprets to message
   *   - other type: ignored
   * @return bool|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   */
  public static function log($exception = NULL, $options = NULL) {
    if (!($a = self::_trc($exception, 'log', $options))) {
      return $a;
    }
    $opts =& $a[0];
    $nl = $opts['newline'];
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag'])) {
      $tagStart = '<' . $u . ' class="module-inspect">';
      $tagEnd = '</' . $u . '>';
    }
    return self::log_db(
        $tagStart
          . ( empty($opts['message']) ? '' : ($opts['message'] . ':' . $nl) )
          . '[Inspect trace - ' . self::$_sessionNo . ':' . self::$_requestNo . ':' . (++self::$_logNo)
              . ' - limit:' . $opts['limit'] . '|depth:' . $opts['depth'] . ']|truncate:' . $opts['truncate'] . ']'
              . $nl
          . $a[1]
          . $tagEnd,
        $opts['category'],
        array_key_exists('severity', $opts) ? $opts['severity'] : -1
    );
  }

  /**
   * Inspect and log stack trace to file, if permitted.
   *
   *   Array/object value of $options are like InspectTrace::log()'s options, except:
   *   - (boolean) by_user: log to user specific file instead of common file
   *
   * @see Inspect::file()
   * @param Exception|falsy $exception
   *   - default: NULL (~ create new backtrace)
   *   - if Exception, trace that
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to limit (maximum stack frame depth)
   *   - string: interprets to message
   *   - other type: ignored
   * @return bool|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   */
  public static function file($exception = NULL, $options = NULL) {
    if (!($a = self::_trc($exception, 'file', $options))) {
      return $a;
    }
    $opts =& $a[0];
    $nl = $opts['newline'];
    return self::log_file(
        ( empty($opts['message']) ? '' : ($opts['message'] . ':' . $nl) )
          . '[Inspect trace - ' . self::$_sessionNo . ':' . self::$_requestNo . ':' . (++self::$_logNo)
              . ' - limit:' . $opts['limit'] . '|depth:' . $opts['depth'] . ']|truncate:' . $opts['truncate'] . ']'
              . $nl
          . $a[1],
        $opts['category'],
        array_key_exists('severity', $opts) ? $opts['severity'] : -1,
        empty($opts['by_user']) ? FALSE : $opts['by_user']
    );
  }

  /**
   * Inspect stack trace and get the output as string, if permitted.
   *
   * @see InspectTrace::log()
   * @param Exception|falsy $exception
   *   - default: NULL (~ create new backtrace)
   *   - if Exception, trace that
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to limit (maximum stack frame depth)
   *   - string: interprets to message
   *   - other type: ignored
   * @return string
   *   - empty if user isnt permitted to get inspections, or tracing failed (other error than exceeding max. output length)
   */
  public static function get($exception = NULL, $options = NULL) {
    if (!($a = self::_trc($exception, 'get', $options))) {
      return '';
    }
    $opts =& $a[0];
    $nl = $opts['newline'];
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag'])) {
      $tagStart = '<' . $u . ' class="module-inspect">';
      $tagEnd = '</' . $u . '>';
    }
    return $tagStart
        . ( empty($opts['message']) ? '' : ($opts['message'] . ':' . $nl) )
        . '[Inspect trace - limit:' . $opts['limit'] . '|depth:' . $opts['depth'] . ']|truncate:' . $opts['truncate'] . ']'
            . $nl
        . $a[1]
        . $tagEnd;
  }
}